from fractions import Fraction
from pygame import Vector2, time, mask, Surface, draw, transform, SRCALPHA, Rect
from math import inf, cos, sin, atan2, radians, degrees, pi
from .features import UILabel
from random import randint


# Define some methods to help with common tasks.
def flatten_list(list_of_lists, flat_list=[]):
    if flat_list == []:
        return flat_list
    else:
        for item in list_of_lists:
            if type(item) == list:
                flatten_list(item)
            else:
                flat_list.append(item)

        return flat_list


def rgb_to_hex(rgb):
    hex = ""
    for value in rgb:
        hex += "%02x" + str(value)
    return hex


def calculate_aspect_ratio(width, height):
    ratio_str = str(Fraction(width, height)).split("/")
    w = int(ratio_str[0])
    h = int(ratio_str[1])
    return w, h


def ccw(a,b,c):
    return (c.y-a.y) * (b.x-a.x) > (b.y-a.y) * (c.x-a.x)


# Return true if line segments AB and CD intersect
def intersect(a, b, c, d):
    return ccw(a, c, d) != ccw(b, c, d) and ccw(a, b, c) != ccw(a, b, d)


def point_in_polygon(pt, poly):
    result = False
    for i in range(len(poly)-1):
        if intersect(Vector2(poly[i].x, poly[i].y), Vector2(poly[i+1].x, poly[i+1].y), Vector2(pt.x, pt.y), Vector2(inf, pt.y)):
            result = not result
    if intersect(Vector2(poly[-1].x, poly[-1].y), Vector2(poly[0].x, poly[0].y), Vector2(pt.x, pt.y), Vector2(inf, pt.y)):
        result = not result
    return result


def closest_point_on_line(line_p1, line_p2, point):
    # Calculates the nearest point on a line from a given point.
    line_p1 = Vector2(line_p1)
    line_p2 = Vector2(line_p2)
    point = Vector2(point)

    A1 = line_p2.y - line_p1.y
    B1 = line_p1.x - line_p2.x
    C1 = (line_p2.y - line_p1.y) * line_p1.x + (line_p1.x - line_p2.x) * line_p1.y
    C2 = -B1*point.x + A1*point.y
    det = A1*A1 - -B1*B1
    cx = 0
    cy = 0
    if det != 0:
        cx = (A1*C1 - B1*C2)/det
        cy = (A1*C2 - -B1*C1)/det
    else:
        cx = point.x
        cy = point.y
    return Vector2(cx, cy)


def render_grid(dest, cols, rows, offset=(0, 0), unit_size=16, line_color=(0, 0, 0)):
    w = cols * unit_size
    h = rows * unit_size

    for y in range(rows):
        pg.draw.line(dest, line_color, (offset[0], offset[1] + y*unit_size),
                     (offset[0] + cols*unit_size, offset[1] + y*unit_size))

    for x in range(cols):
        pg.draw.line(dest, line_color, (offset[0] + x*unit_size, offset[1]),
                     (offset[0] + x*unit_size, offset[1] + rows*unit_size))

    pg.draw.rect(dest, line_color, (0, 0, w, h), 1)


class Timer:
    def __init__(self):
        self.start_time = 0
        self.elapsed_time = 0
        self.started = False

    def start(self):
        self.started = True
        self.start_time = time.get_ticks()

    def stop(self):
        self.started = False
        self.elapsed_time = 0

    def update(self):
        if self.started:
            self.elapsed_time = time.get_ticks() - self.start_time


def draw_tooltip(dest, text, position):
    tooltip_label = UILabel(text)
    tooltip_label.font_size = 12
    border_rect = tooltip_label.surface.get_rect()
    border_rect.x = position[0]
    border_rect.y = position[1]
    draw.rect(dest, (110, 110, 110), border_rect)
    tooltip_label.draw(dest, position)


def render_outline(dest, surface, position, angle=0, color=(255, 255, 255, 255), border_size=1):
    surf_mask = mask.from_surface(surface)
    mask_outline = surf_mask.outline()

    surf = Surface(surface.get_size(), SRCALPHA)
    for point in mask_outline:
        draw.rect(surf, color, (point[0], point[1], border_size, border_size))

    surf = transform.rotate(surf, angle)
    px = position[0]
    py = position[1]
    dest.blit(surf, (px, py))


def smooth_turn(goal_angle, current_angle, turn_speed, dt):
    diff = (goal_angle-current_angle+pi) % (2*pi) - pi
    return current_angle + (diff*turn_speed) * dt


def get_distance(point1, point2):
    point1 = Vector2(point1)
    return point1.distance_to(point2)


def wrap_string(text, max_chars):
    lines = []
    while len(text) > 0:
        count = 0
        text_len = len(text)

        if text_len >= max_chars+1:
            cindex = max_chars - count
            char = text[cindex]

            while char != " ":
                cindex = max_chars - count
                char = text[cindex]
                count += 1

            lines.append(text[0:cindex + 1])
            text = text[(cindex + 1):]
        else:
            lines.append(text)
            text = ""

    return lines


def clamp(value, min_val, max_val):
    if value < min_val: value = min_val
    elif value > max_val: value = max_val
    return value


def lerp(a, b, t):
    return a + (b-a) * t


def randompoint(x_limits, y_limits):
    return randint(x_limits[0], x_limits[1]), randint(y_limits[0], y_limits[1])


def pad_rect(rect, h_pad, v_pad):
    rect = Rect(rect)
    rect.x -= h_pad
    rect.y -= v_pad
    rect.w += h_pad*2
    rect.h += v_pad*2
    return rect


class DataGrid:
    """A grid of data for any purpose."""
    def __init__(self, size):
        self.size = size
        self.data = [0 for x in range(self.size[0]*self.size[1])]
        self.rect = Rect(0, 0, size[0], size[1])

    def load_data(self, map_data):
        # Load in map data using a 2D list of values
        self.size = (len(map_data[0]), len(map_data))

        self.data = [0 for x in range(self.size[0]*self.size[1])]

        for y in range(self.size[1]):
            for x in range(self.size[0]):
                self.set_at((x, y), map_data[y][x])

    def get_area(self, rect, filled=True):
        self.rect = Rect(rect)
        area = [[0 for _ in range(self.rect.h)] for _ in range(self.rect.w)]

        for y in range(0, self.rect.h):
            for x in range(0, self.rect.w):
                area[x][y] = self.get_at((self.rect.x + x, self.rect.y + y))
        return area

    def get_2D_data(self):
        data = []
        for x in range(self.size[0]):
            data.append([])
            for y in range(self.size[1]):
                data[-1].append(self.get_at((x, y)))
        return data

    def get_at(self, coord):
        return self.data[self.size[0] * coord[1] + coord[0]]

    def set_at(self, coord, value):
        i = self.size[0] * coord[1] + coord[0]
        self.data[i] = value

    def print_data(self):
        for y in range(self.size[1]):
            for x in range(self.size[0]):
                print(self.get_at((x, y)), end=' ')
            print()


# Line Intersection
# Calc the gradient 'm' of a line between p1 and p2
def calculate_gradient(p1, p2):
    # Ensure that the line is not vertical
    if (p1[0] != p2[0]):
        m = (p1[1] - p2[1]) / (p1[0] - p2[0])
        return m
    else:
        return None


# Calc the point 'b' where line crosses the Y axis
def calculate_yaxis_intersect(p, m):
    return p[1] - (m * p[0])


# Calc the point where two infinitely long lines (p1 to p2 and p3 to p4) intersect.
# Handle parallel lines and vertical lines (the later has infinate 'm').
# Returns a point tuple of points like this ((x,y),...)  or None
# In non parallel cases the tuple will contain just one point.
# For parallel lines that lay on top of one another the tuple will contain
# all four points of the two lines
def get_intersect_point(p1, p2, p3, p4):
    m1 = calculate_gradient(p1, p2)
    m2 = calculate_gradient(p3, p4)

    # See if the the lines are parallel
    if (m1 != m2):
        # Not parallel

        # See if either line is vertical
        if (m1 is not None and m2 is not None):
            # Neither line vertical
            b1 = calculate_yaxis_intersect(p1, m1)
            b2 = calculate_yaxis_intersect(p3, m2)
            x = (b2 - b1) / (m1 - m2)
            y = (m1 * x) + b1
        else:
            # Line 1 is vertical so use line 2's values
            if (m1 is None):
                b2 = calculate_yaxis_intersect(p3, m2)
                x = p1[0]
                y = (m2 * x) + b2
            # Line 2 is vertical so use line 1's values
            elif (m2 is None):
                b1 = calculate_yaxis_intersect(p1, m1)
                x = p3[0]
                y = (m1 * x) + b1
            else:
                assert False

        return ((x, y),)
    else:
        # Parallel lines with same 'b' value must be the same line so they intersect
        # everywhere in this case we return the start and end points of both lines
        # the calculate_intersect_point method will sort out which of these points
        # lays on both line segments
        b1, b2 = None, None  # vertical lines have no b value
        if m1 is not None:
            b1 = calculate_yaxis_intersect(p1, m1)

        if m2 is not None:
            b2 = calculate_yaxis_intersect(p3, m2)

        # If these parallel lines lay on one another
        if b1 == b2:
            return p1, p2, p3, p4
        else:
            return None


def get_tween_values(start, end, frame_count):
    d = end - start
    dx = d/frame_count
    cval = start
    values = []

    for val in range(frame_count):
        cval += dx
        values.append(cval)
    return values


# For line segments (ie not infinitely long lines) the intersect point
# may not lay on both lines.
#
# If the point where two lines intersect is inside both line's bounding
# rectangles then the lines intersect. Returns intersect point if the line
# intesect o None if not
def calculate_intersect_point(p1, p2, p3, p4):
    p = get_intersect_point(p1, p2, p3, p4)

    if p is not None:
        width = p2[0] - p1[0]
        height = p2[1] - p1[1]
        r1 = Rect(p1, (width, height))
        r1.normalize()

        width = p4[0] - p3[0]
        height = p4[1] - p3[1]
        r2 = Rect(p3, (width, height))
        r2.normalize()

        # Ensure both rects have a width and height of at least 'tolerance' else the
        # collidepoint check of the Rect class will fail as it doesn't include the bottom
        # and right hand side 'pixels' of the rectangle
        tolerance = 1
        if r1.width < tolerance:
            r1.width = tolerance

        if r1.height < tolerance:
            r1.height = tolerance

        if r2.width < tolerance:
            r2.width = tolerance

        if r2.height < tolerance:
            r2.height = tolerance

        for point in p:
            try:
                res1 = r1.collidepoint(point)
                res2 = r2.collidepoint(point)
                if res1 and res2:
                    point = [int(pp) for pp in point]
                    return Vector2(point)

            except ValueError:
                # sometimes the value in a point are too large for PyGame's Rect class
                s = "point was invalid  ", point
                print(s)

        # This is the case where the infinitely long lines crossed but
        # the line segments didn't
        return None

    else:
        return None


def line_box_intersect(p1, p2, rect):
    p1, p2 = Vector2(p1), Vector2(p2)
    rect = Rect(rect)
    tl = Vector2(rect.topleft)

    tr = Vector2(rect.topright)

    bl = Vector2(rect.bottomleft)
    br = Vector2(rect.bottomright)
    points_intersected = []

    if p1.x - rect.left < 0:
        # Left
        left = calculate_intersect_point(p1, p2, tl, bl)
        if left:
            points_intersected.append(left)

    elif p1.x - rect.right > 0:
        # Right
        right = calculate_intersect_point(p1, p2, tr, br)
        if right:
            points_intersected.append(right)

    if p1.y - rect.top < 0:
        # Top
        top = calculate_intersect_point(p1, p2, tl, tr)
        if top:
            points_intersected.append(top)

    elif p1.y - rect.bottom > 0:
        # Bottom
        bottom = calculate_intersect_point(p1, p2, bl, br)
        if bottom:
            points_intersected.append(bottom)

    least_id = 0
    least_d = 100000

    for i, point in enumerate(points_intersected):
        d = Vector2(point).distance_to(p1)
        if d < least_d:
            least_d = d
            least_id = i

    if points_intersected:
        return points_intersected[least_id]


def render_gradient_circle(radius, color1=(255, 217, 58, 150), color2=(255, 255, 255, 100)):
    sw, sh = radius*2, radius*2
    surf = Surface((sw, sh), SRCALPHA)

    for r in range(radius):
        rd = r*1.5
        c = [color1[0]+rd, color1[1]+rd, color1[2]+rd, color1[3]-r*1]
        if c[0] <= 0: c[0] = 0
        if c[1] <= 0: c[1] = 0
        if c[2] <= 0: c[2] = 0
        if c[3] <= 0: c[3] = 0

        if c[0] >= 255: c[0] = 255
        if c[1] >= 255: c[1] = 255
        if c[2] >= 255: c[2] = 255
        if c[3] >= 255: c[3] = 255

        draw.circle(surf, c, (sw/2, sh/2), r, 2)
    return surf


def get_radial_point(center, angle, radius):  # angle is in degrees and converted to radians below.
    p2x = center.x + radius * cos(radians(angle))
    p2y = center.y + radius * sin(radians(angle))
    return Vector2(p2x, p2y)


def angle_to(p1, p2):
    return atan2(p2[1] - p1[1], p2[0] - p1[0])


def draw_arrow(dest, start, end, color=(0, 0, 0)):
    draw.line(dest, color, start, end, 2)
    draw.circle(dest, color, end, 3)


def render_tiled_surface(dest, texture_surface):
    for x in range(int(dest.get_width()/32)):
        for y in range(int(dest.get_height()/32)):
            dest.blit(texture_surface, (x * texture_surface.get_width(), y * texture_surface.get_height()))


def get_poly_bounds(points):
    min_x, max_x = (inf, -inf)
    min_y, max_y = (inf, -inf)
    for point in points:
        if point[0] < min_x:
            min_x = point[0]

        elif point[0] > max_x:
            max_x = point[0]

        if point[1] < min_y:
            min_y = point[1]
        elif point[1] > max_y:
            max_y = point[1]

    min_point = Vector2(min_x, min_y)
    max_point = Vector2(max_x, max_y)
    return min_point, max_point


def get_poly_size(points):
    poly_bounds = get_poly_bounds(points)
    pw = poly_bounds[1][0] - poly_bounds[0][0]
    ph = poly_bounds[1][1] - poly_bounds[0][1]
    return abs(pw), abs(ph)


def generate_textured_polygon(points, texture_surface):
    pw, ph = get_poly_size(points)
    poly_surf = Surface((pw+20, ph+20), SRCALPHA)
    poly_surf.fill((255, 255, 255, 255))
    draw.polygon(poly_surf, (255, 0, 0, 0), points)

    textured_surf = Surface(poly_surf.get_size(), SRCALPHA)
    textured_surf.blit(texture_surface, (0, 0))
    textured_surf.blit(poly_surf, (0, 0))
    textured_surf.set_colorkey((255, 255, 255, 255))
    draw.polygon(textured_surf, (0, 0, 0), points, 3)
    return textured_surf



import pygame as pg
from pygame import Vector2, Rect
from pygame.locals import *
from math import cos, sin, radians
from .codekit import rgb_to_hex
""" Initialize Sub-Systems """

#from pygame.constants import *

# Define some constants
DEBUG_MODE = False
DEFAULT_DISPLAY_SIZE = 1600, 900
COLORS = pg.colordict.THECOLORS


""" 
Define the methods needed to handle most lower level engine features including loading assets, and handling player
input.
"""


def init():
    pg.init()
    pg.font.init()
    pg.joystick.init()
    pg.mixer.init()


def get_events():
    return pg.event.get()


def get_mouse_pos():
    return pg.mouse.get_pos()


def get_keys_pressed():
    return pg.key.get_pressed()


def get_mouse_pressed():
    return pg.mouse.get_pressed(3)


def load_image(fpath, convert=False):
    return pg.image.load(fpath)


def surface_to_surfarray(surface):
    return pg.surfarray.pixels2d(surface.copy())


def surfarray_to_surface(surf_array):
    return pg.surfarray.make_surface(surf_array)


def load_spritesheet(fpath, scale=1, smooth_scaling=False):
    sprite_sheet = load_image(fpath)
    pixel_array = surface_to_surfarray(sprite_sheet)

    vertical_lines = []
    horizontal_lines = []
    border_color_hex = 4294902014

    # Detect horizontal and vertical lines and store their end points in their respective lists.
    # for each horizontal line, check if either end pixel matches an end pixel of a vertical line.
    # if so the length of the horizontal and vertical lines is equivalent to the width and height of that border.
    # Technically, only the top and left edges of each rect need to be detected to find the size of the border.

    # This method of detecting sprite borders may need optimization, but for now it works perfect!
    # Get horizontal border edges.
    for y in range(sprite_sheet.get_height()):
        horizontal_pixels = []
        for x in range(sprite_sheet.get_width()):
            if pixel_array[x, y] == border_color_hex:
                horizontal_pixels.append((x, y))
            else:
                if horizontal_pixels:
                    line = [horizontal_pixels[0], horizontal_pixels[-1]]
                    if not line[0] == line[1]:
                        horizontal_lines.append(line)
                    horizontal_pixels = []

    # Get vertical border edges.
    for x in range(sprite_sheet.get_width()):
        vertical_pixels = []

        for y in range(sprite_sheet.get_height()):
            if pixel_array[x, y] == border_color_hex:
                vertical_pixels.append((x, y))
            else:
                if vertical_pixels:
                    line = [vertical_pixels[0], vertical_pixels[-1]]
                    if not line[0] == line[1]:
                        vertical_lines.append(line)
                    vertical_pixels = []

    border_rects = []
    # Finally we generate the border rects, and append them to a list.
    for hindex in range(len(horizontal_lines)):
        for vindex in range(0, len(vertical_lines)):
            vline = vertical_lines[vindex]
            for point in vline:
                hline = horizontal_lines[hindex]
                if point == hline[0]:
                    rect = Rect(point[0] + 1, point[1] + 1, hline[1][0] - hline[0][0] - 1,
                                     vline[1][1] - vline[0][1] - 1)
                    border_rects.append(rect)
                break

    sprite_surfaces = []
    for rect in border_rects:
        sprite = sprite_sheet.subsurface(rect)
        if smooth_scaling:
            sprite = pg.transform.smoothscale(sprite, (sprite.get_width() * scale, sprite.get_height() * scale))

        else:
            sprite = pg.transform.scale(sprite, (sprite.get_width() * scale, sprite.get_height() * scale))
        sprite_surfaces.append(sprite)

    return sprite_surfaces


def load_sfx(fpath):
    return pg.mixer.Sound(fpath)


def load_music(fpath):
    return pg.mixer.music.load(fpath)


""" Define Core Classes """


class ColorData(Color):
    def __init__(self, r=0, g=0, b=0, a=255):
        super().__init__(r, g, b, a)

    def get_hex(self):
        return rgb_to_hex((self.r, self.g, self.b, self.a))


class Display:
    def __init__(self, size=DEFAULT_DISPLAY_SIZE, caption="Display", flags=0):
        self.surface = pg.display.set_mode(size, flags)
        self._size = size
        self._caption = caption
        self.clock = pg.time.Clock()
        self.frame_rate = 120
        self.dt = 1
        self.flags = flags
        pg.display.set_caption(caption)

    def get_driver(self):
        return pg.display.get_driver()

    def get_wm_info(self):
        return pg.display.get_wm_info()

    def get_info(self):
        return pg.display.Info()

    def get_size(self):
        return self._size

    def get_width(self):
        return self._size[0]

    def get_height(self):
        return self._size[1]

    def get_caption(self):
        return pg.display.get_caption()

    def set_caption(self, caption):
        self._caption = caption
        pg.display.set_caption(caption)

    def set_mode(self, size, flags=0):
        self._size = size
        self.surface = pg.display.set_mode(size, self.flags, vsync=True)

    def clear(self, color=(255, 255, 255, 255)):
         pass

    def blit(self, surface, position):
        pass

    def update(self):
        self.clock.tick(self.frame_rate)


class PGDisplay(Display):
    """ An easy to use Pygame display."""
    def __init__(self, size=DEFAULT_DISPLAY_SIZE, caption="Pygame Display", flags=0):
        super().__init__(size, caption, flags)

    def clear(self, color=(255, 255, 255, 255)):
        self.surface.fill(color)

    def blit(self, surface, position):
        self.surface.blit(surface, position)

    def update(self):
        pg.display.update()
        super().update()


class GLDisplay(Display):
    """ An easy to use OpenGL display."""
    def __init__(self, size=DEFAULT_DISPLAY_SIZE, caption="OpenGL Display", flags=0):
        super().__init__(size, caption, DOUBLEBUF | OPENGL | flags)
        glEnable(GL_DEPTH_TEST)

    def clear(self, color=(255, 255, 255, 255)):
        c = Color(color)
        glClearColor(c.r, c.g, c.b, c.a)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def blit(self, surface, position):
        """
            The blit method for GLDisplay is handled differently than Pygame. Surfaces will likely be treated as
            Textures possibly rendered to some mesh. Study some more OpenGL and come back here to really make progress
            developing this type of Display. I may change the naming of these basic Display methods to better suit the
            hybrid nature of the abstracted Display class, so that minimal to no code changes are necessary to switch
            between a PGDisplay and a GLDisplay.

            At first any rendering of 3D graphics and meshes will only be available on a GLDisplay, but it's possible
            to eventually implement a rudimentary 3D rendering pipeline to use in any case that OpenGL is not available
            or practical. PGDisplay is essentially a software rendering pipeline, while OpenGL implements a much
            more powerful hardware accelerated rendering pipeline. This is a common feature for many graphical
            applications, so it will likely be useful in some cases.
        """
        pass

    def draw_cube(self, position=(0, 0, 0), color=(255, 255, 255), size=1, wire=False):
        glColor3f(color[0], color[1], color[2])
        glPushMatrix()
        glTranslatef(position[0], position[1], position[2])

        if wire:
            glutWireCube(size)
        else:
            glutSolidCube(size)
        glPopMatrix()

    def draw_sphere(self, position=(0, 0, 0), color=(255, 255, 255), size=1, slices=(30, 30), wire=False):
        glColor3f(color[0], color[1], color[2])
        glPushMatrix()
        glTranslatef(position[0], position[1], position[2])
        # size is 2x the radius, so we half the given size.
        if wire:
            glutWireSphere(size * .5, slices[0], slices[1])
        else:
            glutSolidSphere(size * .5, slices[0], slices[1])
        glPopMatrix()

    def update(self):
        pg.display.flip()
        #super().update()





""" UI Features """
from . import core
from . import codekit as ck
from math import cos, sin, radians, degrees, atan2
core.pg.font.init()
DEFAULT_FONT = "Arial"


class Component:
    current_cid = 0

    def __init__(self):
        # Each component will link to the parent GameObject for easy reference and access.
        self.cid = Component.current_cid
        Component.current_cid += 1
        self.type = None

    def on_init(self):
        pass

    def on_event(self, events, mouse_pos):
        pass

    def on_update(self, dt):
        pass

    def on_draw(self, dest, offset=(0, 0)):
        pass


class Transform(Component):
    def __init__(self, position=(0, 0), rotation=0):
        super().__init__()
        self.position = core.Vector2(position)
        self.local_position = core.Vector2(0, 0)
        self.rotation = rotation
        self.local_rotation = 0
        self.scale = 1
        self.local_scale = 1

        self.old_position = self.position
        self.old_rotation = self.rotation
        self.old_scale = self.scale
        self._has_changed = False

    def translate(self, translation, relative_to=None):
        pass


class GameObject:
    current_oid = 0

    def __init__(self, name, position=(0, 0), rotation=0, parent=None, surface=None):
        self.oid = GameObject.current_oid
        GameObject.current_oid += 1
        self.name = name

        self.parent = parent
        self.transform = Transform(position, rotation)

        self.components = [self.transform]

        # If the object has a parent we need to calculate the initial local offset between the parent and the new object
        if self.parent:
            self.transform.local_position = self.transform.position
            self.d = self.parent.transform.position.distance_to(self.transform.local_position)

        bound_size = (32, 32)
        self.rect = core.Rect(self.transform.position.x - bound_size[0]*2,
                         self.transform.position.y - bound_size[1]*2, bound_size[0], bound_size[1])

        if surface:
            self.surface = surface
            bound_size = self.surface.get_size()
            self.rect = core.Rect(self.transform.position.x - bound_size[0] * 2,
                             self.transform.position.y - bound_size[1] * 2, bound_size[0], bound_size[1])
        else:
            self.surface = core.pg.Surface(self.rect.size)

        self.scaled_surface = self.surface.copy()
        self.transformed_surface = self.surface.copy()

    def on_init(self):
        for component in self.components:
            component.on_init()

    def on_event(self, events, mouse_pos):
        for component in self.components:
            component.on_event(events, mouse_pos)

    def on_update(self, dt):
        for component in self.components:
            component.on_update(dt)

        if self.parent:
            if self.parent.transform.old_position != self.parent.transform.position:
                self.d = self.parent.transform.position.distance_to(self.transform.local_position)

            self.transform.rotation = self.parent.transform.rotation + self.transform.local_rotation

            # # TEMPORARY
            # # END TEMPORARY

            r = -radians(self.parent.transform.rotation)
            ox = self.d * cos(r)
            oy = self.d * sin(r)

            self.transform.position = self.parent.transform.position + (ox, oy)

        # Any time a change has been made to the transform, other than positional ones, will require both scaling
        # and rotation to be re-applied
        # Check if the transform has changed in scale, position, or rotation, and if so update the object surface.

        # Position
        if self.transform.position != self.transform.old_position:
            self.transform.old_position = self.transform.position

        # Rotation and Scale
        if (self.transform.old_scale != self.transform.scale) or (self.transform.old_rotation != self.transform.rotation):
            # Scale
            if self.transform.old_scale != self.transform.scale:
                if self.transform.scale < .1:
                    self.transform.scale = .1
                self.transform.old_scale = self.transform.scale

            self.scaled_surface = core.pg.transform.scale(self.surface, (self.surface.get_width() * self.transform.scale,
                                                                    self.surface.get_height() * self.transform.scale))

            # Rotate
            if self.transform.old_rotation != self.transform.rotation:
                self.transform.old_rotation = self.transform.rotation

            self.transformed_surface = core.pg.transform.rotate(self.scaled_surface, self.transform.rotation)
            self.transform.old_rotation = self.transform.rotation

        # Update object's rect
        self.rect = core.Rect(self.transform.position.x - self.transformed_surface.get_width() * 2,
                         self.transform.position.y - self.transformed_surface.get_height() * 2,
                         self.transformed_surface.get_width(), self.transformed_surface.get_height())

        self.rect.center = self.transform.position

    def on_draw(self, dest, offset=(0, 0)):
        for component in self.components:
            component.on_draw(dest, offset)

        # If the object has a parent, draw it relative to the parent's transform.
        # Otherwise, just draw the object at its global position.
        # Or should all objects simply be placed using global coordinates, and then we can apply the relative offset.

        dest.blit(self.transformed_surface, self.rect.center - core.Vector2(self.rect.w/2, self.rect.h/2))

        if core.DEBUG_MODE:
            if self.parent:
                core.pg.draw.line(dest, (0, 0, 0), self.transform.position, self.parent.transform.position)
                core.pg.draw.circle(dest, (0, 0, 255), self.transform.position, 3)


class UIElement:
    def __init__(self):
        self.surface = core.pg.Surface((32, 32), core.SRCALPHA)
        self.rect = core.Rect(0, 0, self.surface.get_width(), self.surface.get_height())
        self.position = core.Vector2(self.rect.topleft)

    def on_event(self, events, mouse_pos):
        pass

    def on_update(self, dt):
        self.position = core.Vector2(self.rect.topleft)

    def on_draw(self, dest):
        pass


class UILabel(UIElement):
    def __init__(self, text, position=(0, 0), font_size=20, font_color=(0, 0, 0), bg_color=None):
        super().__init__()
        self._font_size = font_size
        self._font_color = font_color
        self.bg_color = bg_color
        self._text = text
        self._font_path = None

        self.font = core.pg.font.SysFont(DEFAULT_FONT, font_size)
        self.surface = self.font.render(self._text, True, self._font_color, self.bg_color)
        self.rect = core.Rect(position[0], position[1], self.surface.get_width(), self.surface.get_height())

    def set_sys_font(self, font_name=None, font_size=None):
        if not font_size:
            font_size = self._font_size
        else:
            self._font_size = font_size

        self.font = core.pg.font.SysFont(font_name, font_size)
        self.surface = self.font.render(self.text, True, self._font_color, self.bg_color)
        self.rect = core.Rect(self.rect.x, self.rect.y, self.surface.get_width(), self.surface.get_height())

    def set_ttf_font(self, font_path=None, font_size=None):
        if not font_path:
            font_path = self._font_path
        else:
            self._font_path = font_path

        if not font_size:
            font_size = self._font_size
        else:
            self._font_size = font_size

        self.font = core.pg.font.Font(font_path, font_size)
        self.surface = self.font.render(self.text, True, self._font_color, self.bg_color)
        self.rect = core.Rect(self.rect.x, self.rect.y, self.surface.get_width(), self.surface.get_height())

    def set_bold(self, value=True):
        self.font.set_bold(value)
        self.surface = self.font.render(self._text, True, self._font_color, self.bg_color)

    def set_underline(self, value=True):
        self.font.set_underline(value)
        self.surface = self.font.render(self._text, True, self._font_color, self.bg_color)

    @property
    def font_color(self, color):
        self._font_color = color
        self.surface = self.font.render(self.text, True, self._font_color, self.bg_color)

    @font_color.getter
    def font_color(self):
        return self._font_color

    @font_color.setter
    def font_color(self, color):
        self._font_color = color
        self.surface = self.font.render(self.text, True, self._font_color, self.bg_color)

    @property
    def text(self, value):
        self._text = value
        self.surface = self.font.render(self._text, True, self._font_color, None)
        self.rect = core.Rect(self.rect.x, self.rect.y, self.surface.get_width(), self.surface.get_height())

    @text.getter
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        self.surface = self.font.render(self._text, True, self._font_color, (0, 0, 0))
        self.rect = core.Rect(self.rect.x, self.rect.y, self.surface.get_width(), self.surface.get_height())

    def on_draw(self, dest, centerx=False):
        if centerx:
            dest.blit(self.surface, (self.rect.centerx - self.surface.get_width()/2, self.rect.centery - self.surface.get_height()/2))
        else:
            dest.blit(self.surface, self.rect.topleft)

        if core.DEBUG_MODE:
            core.pg.draw.rect(dest, (255, 0, 0), self.rect, 1)


class UILink(UILabel):
    def __init__(self, text, position=(0, 0), font_size=20, font_color=(0, 0, 0), bg_color=None, command=None, cmd_args=None):
        super().__init__(text, position, font_size, font_color, bg_color)
        self.mouseover = False
        self.clicked = False
        self.right_clicked = False
        self.holding = False
        self.released = False
        self.command = command
        self.args = cmd_args

        self.base_color = [val for val in font_color]
        self.hover_color = (0, 0, 255)
        self.clicked_color = (255, 0, 0)

    def on_event(self, events, mouse_pos):
        self.mouseover = False
        self.clicked = False
        self.right_clicked = False
        self.released = False

        if self.rect.collidepoint(mouse_pos):
            self.mouseover = True

        for event in events:
            if event.type == core.MOUSEBUTTONDOWN:
                if event.button == 1:
                    if self.mouseover:
                        self.clicked = True
                        self.holding = True

            if event.type == core.MOUSEBUTTONUP:
                if event.button == 1:
                    if self.mouseover:
                        if self.holding:
                            self.clicked = False
                            if self.command:
                                if self.args:
                                    self.command(self.args)
                                else:
                                    self.command()

                    self.holding = False
                    self.released = True
                elif event.button == 3:
                    if self.mouseover:
                        self.right_clicked = True

        if core.pg.mouse.get_focused():
            if self.holding and self.clicked_color:
                self.font_color = self.clicked_color
            elif self.mouseover and self.hover_color:
                self.font_color = self.hover_color
            else:
                self.font_color = self.base_color
        else:
            self.font_color = self.base_color


class UIButton(UILink):
    def __init__(self, text, position=(0, 0), font_size=20, font_color=(0, 0, 0), bg_color=None, command=None, cmd_args=()):
        super().__init__(text, position, font_size, font_color, bg_color, command, cmd_args)
        self.rect.w += 10
        self.rect.h += 10
        self.clicked_color = (255, 255, 255)
        self.hover_color = None

    def on_draw(self, dest, centerx=True):
        if self.mouseover and core.pg.mouse.get_focused():
            core.pg.draw.rect(dest, (180, 180, 180), self.rect)

        super().on_draw(dest, centerx)


class UIDropMenu(UIElement):
    def __init__(self, position=(0, 0)):
        super().__init__()
        self.buttons = []
        self.rect.x = position[0]
        self.rect.y = position[1]
        self.font_size = 16
        self.active = False

    def show(self):
        self.active = True

    def hide(self):
        self.active = False
        for button in self.buttons:
            button.holding = False

    def add_option(self, text, command=None, cmd_args=()):
        new_button = UIButton(text, font_size=self.font_size, command=command, cmd_args=cmd_args)
        self.buttons.append(new_button)

        if new_button.rect.w > self.rect.w:
            self.rect.w = new_button.rect.w

        self.rect.h = 0
        for button in self.buttons:
            button.rect.topleft = (self.rect.x, self.rect.y + self.rect.h)
            self.rect.h += button.rect.h
            button.rect.w = self.rect.w

    def on_event(self, events, mouse_pos):
        if self.active:
            for button in self.buttons:
                button.on_event(events, mouse_pos)

            for event in events:
                if event.type == core.MOUSEBUTTONDOWN:
                    if event.button == 1:
                        if self.rect.collidepoint(mouse_pos):
                            for button in self.buttons:
                                if button.clicked:
                                    if button.command:
                                        if button.args:
                                            button.command(button.args)
                                        else:
                                            button.command()

                            self.active = False

    def on_update(self, dt):
        if self.active:
            super().on_update(dt)
            yoff = 0
            for button in self.buttons:
                button.rect.topleft = (self.rect.x, self.rect.y + yoff)
                yoff += button.rect.h

                if button.holding:
                    if button.released:
                        if button.command:
                            if button.args:
                                button.command(button.args)
                            else:
                                button.command()

    def on_draw(self, dest):
        if self.active:
            core.pg.draw.rect(dest, (200, 200, 200), self.rect)
            for button in self.buttons:
                button.on_draw(dest)


class UIMenuBar(UIElement):
    def __init__(self):
        super().__init__()
        self.font_size = 16
        self.options = []

    def add_option(self, text, menu=None):
        if menu:
            button = UIButton(text, font_size=self.font_size, command=menu.show)
        else:
            button = UIButton(text, font_size=self.font_size)

        self.options.append({"button": button, "menu": menu})

        xoff = 0
        for option in self.options:
            option["button"].rect.topleft = (xoff, 0)
            xoff += option["button"].rect.w

        if menu:
            menu.rect.topleft = core.Vector2(option["button"].rect.x, option["button"].rect.y + option["button"].rect.h)

    def on_event(self, events, mouse_pos):
        menu_open = False

        # When the user clicks anywhere within a menu rect, check if a button was clicked. If so, perform the button's
        # command. Then close the menu, whether a button was clicked or not.


        for option in self.options:
            option["button"].on_event(events, mouse_pos)

            if option["menu"]:
                if option["menu"].active:
                    menu_open = True

                # Here we are looping through all the options again to see if a menu is already open. If that is true
                # then any menubar option that the mouse hovers over will open, without the need to be clicked.
                for o in self.options:
                    if o["menu"]:
                        if o["menu"].active:
                            menu_open = True

                if menu_open and option["button"].mouseover:
                    option["menu"].show()

                option["menu"].on_event(events, mouse_pos)

                # While a menu is open, test to make sure the cursor is still within the menu's rect. If not, hide the
                # menu.
                if option["menu"].active:
                    if not ck.pad_rect(option["menu"].rect, 8, 28).collidepoint(mouse_pos):
                        option["menu"].hide()

                    for option2 in self.options:
                        if option2["menu"]:
                            if not option2 == option:
                                option2["menu"].active = False

    def on_update(self, dt):
        for option in self.options:
            option["button"].on_update(dt)
            if option["menu"]:
                option["menu"].on_update(dt)

    def on_draw(self, dest):
        self.rect = core.Rect(0, 0, dest.get_width(), 28)
        core.pg.draw.rect(dest, (200, 200, 200), self.rect)

        for option in self.options:
            option["button"].on_draw(dest)

            if option["menu"]:
                if option["menu"].active:
                    option["menu"].on_draw(dest)
                #core.pg.draw.rect(dest, (255, 0, 0), option["menu"].rect, 1)

        core.pg.draw.line(dest, (0, 0, 0), (0, 28), (dest.get_width(), 28))
